@@ -5,6 +5,7 @@ import sys
 import os
 import signal
 import subprocess
+import time
 from collections import defaultdict
 from inotify_simple import INotify, flags
 
@@ -14,24 +15,34 @@ def _wait_files_creation(file_list):
   # Establish a list of directory and subfiles.
   # and test existence before watching, so that we don't miss an event.
   directories = defaultdict(dict)
-  for f in file_list:
-    dirname, filename = os.path.split(f)
-    directories[dirname][filename] = os.path.lexists(f)
+  def check_if_files_exists():
+    for f in file_list:
+      dirname, filename = os.path.split(f)
+      directories[dirname][filename] = os.path.lexists(f)
+  check_if_files_exists()
 
   def all_files_exists():
     return all(all(six.itervalues(files)) for files in six.itervalues(directories))
 
   with INotify() as inotify:
-    watchdescriptors = {inotify.add_watch(dirname,
-        flags.CREATE | flags.DELETE | flags.MOVED_TO | flags.MOVED_FROM
-        ): dirname
-      for dirname in directories}
-
-    while not all_files_exists():
-      for event in inotify.read():
-        directory = directories[watchdescriptors[event.wd]]
-        if event.name in directory:
-          directory[event.name] = event.mask & (flags.CREATE | flags.MOVED_TO)
+    try:
+      watchdescriptors = {inotify.add_watch(dirname,
+          flags.CREATE | flags.DELETE | flags.MOVED_TO | flags.MOVED_FROM
+          ): dirname
+        for dirname in directories}
+    except OSError as e:
+      if e.errno not in (errno.ENOSPC, errno.EMFILE):
+        raise
+      print('Error using inotify, falling back to polling')
+      while not all_files_exists():
+        time.sleep(0.1)
+        check_if_files_exists()
+    else:
+      while not all_files_exists():
+        for event in inotify.read():
+          directory = directories[watchdescriptors[event.wd]]
+          if event.name in directory:
+            directory[event.name] = event.mask & (flags.CREATE | flags.MOVED_TO)
 
 def _libc():
   from ctypes import CDLL, get_errno, c_char_p, c_int, c_ulong, util
@@ -67,7 +78,7 @@ def generic_exec(args, extra_environ=None, wait_list=None,
     else:
       # With chained shebangs, several paths may be inserted at the beginning.
       n = len(args)
-      for i in xrange(1+len(running)-n):
+      for i in six.moves.xrange(1+len(running)-n):
         if args == running[i:n+i]:
           sys.exit("Already running with pid %s." % pid)
     with open(pidfile, 'w') as f:
@@ -91,16 +102,19 @@ def generic_exec(args, extra_environ=None, wait_list=None,
     uid = os.getuid()
     gid = os.getgid()
     unshare(CLONE_NEWUSER |CLONE_NEWNS)
-    with open('/proc/self/setgroups', 'wb') as f: f.write('deny')
-    with open('/proc/self/uid_map',   'wb') as f: f.write('%s %s 1' % (uid, uid))
-    with open('/proc/self/gid_map',   'wb') as f: f.write('%s %s 1' % (gid, gid))
+    with open('/proc/self/setgroups', 'w') as f:
+      f.write('deny')
+    with open('/proc/self/uid_map', 'w') as f:
+      f.write('%s %s 1' % (uid, uid))
+    with open('/proc/self/gid_map', 'w') as f:
+      f.write('%s %s 1' % (gid, gid))
     for size, path in private_tmpfs:
       try:
         os.mkdir(path)
       except OSError as e:
         if e.errno != errno.EEXIST:
           raise
-      mount('tmpfs', path, 'tmpfs', 0, 'size=' + size)
+      mount(b'tmpfs', path.encode(), b'tmpfs', 0, ('size=' + size).encode())
 
   if extra_environ:
     env = os.environ.copy()
